"""
Общие фикстуры для всех тестов
"""
import pytest
import asyncio
from cachka import cache_registry, CacheConfig


@pytest.fixture(scope="function")
def cache_config():
    """Базовая конфигурация кэша для тестов"""
    return CacheConfig(
        db_path=":memory:",
        vacuum_interval=None,
        cleanup_on_start=False,
        enable_metrics=False,
        enable_encryption=False
    )


@pytest.fixture(scope="function")
async def initialized_cache(cache_config):
    """Инициализированный кэш для тестов"""
    # Сбрасываем перед инициализацией
    if cache_registry.is_initialized():
        try:
            await cache_registry.shutdown()
        except:
            pass
    
    cache_registry.initialize(cache_config)
    yield cache_registry.get()
    
    # Cleanup после теста
    if cache_registry.is_initialized():
        try:
            await cache_registry.shutdown()
        except:
            pass


@pytest.fixture(scope="function", autouse=True)
async def reset_cache_registry_after_test():
    """Автоматически сбрасывает registry после каждого теста"""
    yield
    # Cleanup после теста
    if cache_registry.is_initialized():
        try:
            await cache_registry.shutdown()
        except:
            pass
        # Сбрасываем состояние
        cache_registry.reset()




import time
import threading
import asyncio
from collections import OrderedDict
from logging import getLogger
from typing import Any, Optional, Callable, Awaitable
from contextlib import contextmanager, asynccontextmanager

logger = getLogger(__name__)


class TTLLRUCache:
    def __init__(
        self,
        maxsize: int = 1024,
        ttl: int = 300,
        shards: int = 8,
        enable_metrics: bool = False,
        name: str = "l1_cache",
    ):
        if maxsize <= 0:
            raise ValueError("maxsize must be > 0")
        if ttl <= 0:
            raise ValueError("ttl must be > 0")
        if shards <= 0:
            raise ValueError("shards must be > 0")

        self.maxsize_per_shard = maxsize // shards + 1
        self.ttl = ttl
        self.shards = shards
        self.name = name
        self._is_async = False
        self._bg_task: Optional[asyncio.Task] = None

        # Create shards
        self._shards = [
            _LRUTTLShard(self.maxsize_per_shard, ttl, enable_metrics, name, i)
            for i in range(shards)
        ]

        # Detect async context on first use
        self._determine_mode()

    def _determine_mode(self):
        """Determine if we're in async context (best-effort)."""
        try:
            asyncio.get_running_loop()
            self._is_async = True
        except RuntimeError:
            self._is_async = False

    def _get_shard(self, key: str) -> '_LRUTTLShard':
        """Consistent shard selection."""
        return self._shards[hash(key) % self.shards]

    # === Sync API ===
    def get(self, key: str, default: Any = None) -> Any:
        shard = self._get_shard(key)
        return shard.get(key, default)

    def set(self, key: str, value: Any) -> None:
        shard = self._get_shard(key)
        shard.set(key, value)

    def delete(self, key: str) -> None:
        shard = self._get_shard(key)
        shard.delete(key)

    def touch(self, key: str) -> bool:
        """Extend TTL if exists."""
        shard = self._get_shard(key)
        return shard.touch(key)

    def get_or_set(self, key: str, factory: Callable[[], Any]) -> Any:
        """Atomic get-or-set (thread-safe)."""
        shard = self._get_shard(key)
        return shard.get_or_set(key, factory)

    def cleanup(self) -> int:
        """Force cleanup expired entries across all shards."""
        total = 0
        for shard in self._shards:
            total += shard.cleanup()
        return total

    def __len__(self) -> int:
        return sum(len(shard) for shard in self._shards)

    # === Async API ===
    async def get_async(self, key: str, default: Any = None) -> Any:
        shard = self._get_shard(key)
        return await shard.get_async(key, default)

    async def set_async(self, key: str, value: Any) -> None:
        shard = self._get_shard(key)
        await shard.set_async(key, value)

    async def delete_async(self, key: str) -> None:
        shard = self._get_shard(key)
        await shard.delete_async(key)

    async def touch_async(self, key: str) -> bool:
        shard = self._get_shard(key)
        return await shard.touch_async(key)

    async def get_or_set_async(
        self, key: str, factory: Callable[[], Awaitable[Any]]
    ) -> Any:
        shard = self._get_shard(key)
        return await shard.get_or_set_async(key, factory)

    async def start_background_cleanup(self, interval: int = 60):
        """Start async background cleanup task."""
        if not self._is_async:
            raise RuntimeError("Background cleanup requires async context")
        if self._bg_task is not None:
            raise RuntimeError("Background cleanup already started")

        async def _cleanup_loop():
            while True:
                try:
                    await asyncio.sleep(interval)
                    await asyncio.get_running_loop().run_in_executor(
                        None, self.cleanup
                    )
                except asyncio.CancelledError:
                    break
                except Exception:
                    pass  # log in real app

        self._bg_task = asyncio.create_task(_cleanup_loop())

    async def stop_background_cleanup(self):
        """Stop background cleanup task."""
        if self._bg_task:
            self._bg_task.cancel()
            try:
                await self._bg_task
            except asyncio.CancelledError:
                pass
            self._bg_task = None

    def __getitem__(self, key, default=None):
        return self.get(key=key, default=default)

    def __setitem__(self, key, value):
        self.set(key=key, value=value)

    def __contains__(self, key: str) -> bool:
        shard = self._get_shard(key)
        return key in shard


class _LRUTTLShard:
    """Internal shard with its own lock and cache."""

    def __init__(self, maxsize: int, ttl: int, enable_metrics: bool, cache_name: str, shard_id: int):
        self.maxsize = maxsize
        self.ttl = ttl
        self._cache = OrderedDict()
        self._lock = threading.RLock()
        self._async_lock = None  # lazy init

        # Metrics (optional)
        self._metrics = None
        if enable_metrics:
            self._init_metrics(cache_name, shard_id)

    def _init_metrics(self, cache_name: str, shard_id: int):
        try:
            from prometheus_client import Counter, Histogram
            self._metrics = {
                "hits": Counter(
                    "cache_l1_hits_total",
                    "L1 cache hits",
                    ["cache", "shard"]
                ).labels(cache=cache_name, shard=str(shard_id)),
                "misses": Counter(
                    "cache_l1_misses_total",
                    "L1 cache misses",
                    ["cache", "shard"]
                ).labels(cache=cache_name, shard=str(shard_id)),
                "evictions": Counter(
                    "cache_l1_evictions_total",
                    "L1 cache evictions",
                    ["cache", "shard"]
                ).labels(cache=cache_name, shard=str(shard_id)),
            }
        except ImportError:
            self._metrics = None

    def _now(self) -> float:
        return time.time()

    def _is_expired(self, timestamp: float) -> bool:
        return self._now() - timestamp > self.ttl

    def _cleanup_expired(self) -> int:
        """Remove expired from head of LRU list."""
        removed = 0
        keys_to_remove = []
        for key, (_, ts) in self._cache.items():
            if self._is_expired(ts):
                keys_to_remove.append(key)
                removed += 1
            else:
                break  # OrderedDict is LRU-ordered
        for key in keys_to_remove:
            del self._cache[key]
        return removed

    @contextmanager
    def _sync_lock(self):
        with self._lock:
            yield

    @asynccontextmanager
    async def _async_lock_ctx(self):
        if self._async_lock is None:
            self._async_lock = asyncio.Lock()
        async with self._async_lock:
            yield

    # === Sync methods ===
    def get(self, key: str, default: Any = None) -> Any:
        with self._sync_lock():
            self._cleanup_expired()
            if key in self._cache:
                value, ts = self._cache[key]
                if self._is_expired(ts):
                    del self._cache[key]
                    if self._metrics:
                        self._metrics["misses"].inc()
                    return default
                self._cache.move_to_end(key)
                if self._metrics:
                    self._metrics["hits"].inc()
                return value
            if self._metrics:
                self._metrics["misses"].inc()
            return default

    def set(self, key: str, value: Any) -> None:
        with self._sync_lock():
            now = self._now()
            self._cache[key] = (value, now)
            self._cache.move_to_end(key)
            evicted = 0
            while len(self._cache) > self.maxsize:
                self._cache.popitem(last=False)
                evicted += 1
            if evicted and self._metrics:
                self._metrics["evictions"].inc(evicted)

    def delete(self, key: str) -> None:
        with self._sync_lock():
            self._cache.pop(key, None)

    def touch(self, key: str) -> bool:
        with self._sync_lock():
            if key in self._cache:
                value, _ = self._cache[key]
                self._cache[key] = (value, self._now())
                self._cache.move_to_end(key)
                return True
            return False

    def get_or_set(self, key: str, factory: Callable[[], Any]) -> Any:
        with self._sync_lock():
            self._cleanup_expired()
            if key in self._cache:
                value, ts = self._cache[key]
                if not self._is_expired(ts):
                    self._cache.move_to_end(key)
                    if self._metrics:
                        self._metrics["hits"].inc()
                    return value
            # Miss
            value = factory()
            self.set(key, value)
            if self._metrics:
                self._metrics["misses"].inc()
            return value

    def cleanup(self) -> int:
        with self._sync_lock():
            return self._cleanup_expired()

    def __len__(self) -> int:
        with self._sync_lock():
            self._cleanup_expired()
            return len(self._cache)

    def __contains__(self, key: str) -> bool:
        with self._sync_lock():
            return key in self._cache

    # === Async wrappers ===
    async def get_async(self, key: str, default: Any = None) -> Any:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self.get, key, default)

    async def set_async(self, key: str, value: Any) -> None:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self.set, key, value)

    async def delete_async(self, key: str) -> None:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self.delete, key)

    async def touch_async(self, key: str) -> bool:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self.touch, key)

    async def get_or_set_async(
        self, key: str, factory: Callable[[], Awaitable[Any]]
    ) -> Any:
        # First, try to get
        existing = await self.get_async(key)
        if existing is not None:
            return existing
        # Compute in async
        value = await factory()
        await self.set_async(key, value)
        return value


import pytest
from cachka.utils import make_cache_key


class TestMakeCacheKey:
    """Тесты генерации ключей кэша"""

    def test_key_generation_deterministic(self):
        """Ключ должен быть детерминированным"""
        key1 = make_cache_key("func", (1, 2), {"a": 3})
        key2 = make_cache_key("func", (1, 2), {"a": 3})
        assert key1 == key2

    def test_key_different_for_different_args(self):
        """Разные аргументы = разные ключи"""
        key1 = make_cache_key("func", (1,), {})
        key2 = make_cache_key("func", (2,), {})
        assert key1 != key2

    def test_key_same_for_same_args(self):
        """Одинаковые аргументы = одинаковый ключ"""
        key1 = make_cache_key("func", (1, 2, 3), {"x": 10, "y": 20})
        key2 = make_cache_key("func", (1, 2, 3), {"x": 10, "y": 20})
        assert key1 == key2

    def test_key_handles_kwargs_order(self):
        """Порядок kwargs не важен"""
        key1 = make_cache_key("func", (), {"a": 1, "b": 2})
        key2 = make_cache_key("func", (), {"b": 2, "a": 1})
        assert key1 == key2  # sorted(kwargs.items()) делает порядок неважным

    def test_key_different_functions(self):
        """Разные функции = разные ключи"""
        key1 = make_cache_key("func1", (1,), {})
        key2 = make_cache_key("func2", (1,), {})
        assert key1 != key2

    def test_key_with_none(self):
        """Ключ с None значениями"""
        key1 = make_cache_key("func", (None,), {"x": None})
        key2 = make_cache_key("func", (None,), {"x": None})
        assert key1 == key2

    def test_key_with_complex_objects(self):
        """Сложные объекты в аргументах"""
        obj1 = {"nested": [1, 2, 3]}
        obj2 = {"nested": [1, 2, 3]}
        key1 = make_cache_key("func", (obj1,), {})
        key2 = make_cache_key("func", (obj2,), {})
        assert key1 == key2

    def test_key_with_unicode(self):
        """Unicode в аргументах"""
        key1 = make_cache_key("func", ("тест",), {"ключ": "значение"})
        key2 = make_cache_key("func", ("тест",), {"ключ": "значение"})
        assert key1 == key2

    def test_key_length(self):
        """Длина ключа (SHA256 = 64 hex chars)"""
        key = make_cache_key("func", (1, 2, 3), {"a": "b"})
        assert len(key) == 64  # SHA256 hex digest length
        assert all(c in '0123456789abcdef' for c in key)

    def test_key_empty_args(self):
        """Пустые аргументы"""
        key1 = make_cache_key("func", (), {})
        key2 = make_cache_key("func", (), {})
        assert key1 == key2

    def test_key_only_kwargs(self):
        """Только kwargs"""
        key1 = make_cache_key("func", (), {"x": 1, "y": 2})
        key2 = make_cache_key("func", (), {"y": 2, "x": 1})
        assert key1 == key2

    def test_key_only_args(self):
        """Только args"""
        key1 = make_cache_key("func", (1, 2, 3), {})
        key2 = make_cache_key("func", (1, 2, 3), {})
        assert key1 == key2




import asyncio
import inspect
import functools
from typing import Callable

from logging import getLogger

from .registry import cache_registry
from .core import CacheConfig
from .utils import make_cache_key


logger = getLogger(__name__)


def cached(ttl: int = 300, ignore_self: bool = False):
    def decorator(func: Callable):
        if inspect.iscoroutinefunction(func):
            # Async function
            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                cache = cache_registry.get()
                key_args = args[1:] if ignore_self and args else args
                key = make_cache_key(func.__name__, key_args, kwargs)
                cached_val = await cache.get(key)
                if cached_val is not None:
                    return cached_val
                result = await func(*args, **kwargs)
                await cache.set(key, result, ttl)
                return result
            return wrapper

        else:
            # Sync function
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                cache = cache_registry.get()
                key_args = args[1:] if ignore_self and args else args
                key = make_cache_key(func.__name__, key_args, kwargs)

                # Try L1
                l1_val = None
                with cache.l1_lock:
                    if key in cache.l1_cache:
                        l1_val = cache.l1_cache[key]
                if l1_val is not None:
                    return l1_val

                # Get or create event loop for async operations
                try:
                    loop = asyncio.get_running_loop()
                    # Running loop exists, use run_coroutine_threadsafe
                    future = asyncio.run_coroutine_threadsafe(cache.get(key), loop)
                    cached_val = future.result()
                    
                    if cached_val is not None:
                        return cached_val
                    
                    result = func(*args, **kwargs)
                    
                    future = asyncio.run_coroutine_threadsafe(cache.set(key, result, ttl), loop)
                    future.result()
                except RuntimeError:
                    # No running loop, use asyncio.run for isolated execution
                    async def _get_cached():
                        return await cache.get(key)
                    
                    cached_val = asyncio.run(_get_cached())
                    
                    if cached_val is not None:
                        return cached_val
                    
                    result = func(*args, **kwargs)
                    
                    async def _set_cached():
                        await cache.set(key, result, ttl)
                    
                    asyncio.run(_set_cached())

                return result
            return wrapper
    return decorator


__all__ = ["cached", "cache_registry", "CacheConfig"]


import asyncio
from contextlib import asynccontextmanager

import aiosqlite
import pickle
import time
import os
import threading
import secrets
from typing import Any, Optional, Dict
from abc import ABC, abstractmethod

from cachka.ttllrucache import TTLLRUCache

# === Optional deps ===
try:
    from prometheus_client import generate_latest
except ImportError:
    def generate_latest():
        return b"# Prometheus not available\n"

try:
    from opentelemetry import trace
    from opentelemetry.trace import SpanKind
    HAS_OTEL = True
except ImportError:
    HAS_OTEL = False
    class DummyTracer:
        def start_as_current_span(self, *a, **kw): return self
        def __enter__(self): return self
        def __exit__(self, *a): pass
    trace = type('trace', (), {'get_tracer': lambda *a: DummyTracer()})

try:
    from cryptography.hazmat.primitives.ciphers.aead import AESGCM
    HAS_CRYPTO = True
except ImportError:
    HAS_CRYPTO = False

import structlog
logger = structlog.get_logger(__name__)

# === Config ===
from pydantic import BaseModel

class CacheConfig(BaseModel):
    db_path: str = "cache.db"
    name: str = "default"
    l1_maxsize: int = 1024
    l1_ttl: int = 300
    vacuum_interval: Optional[int] = 3600
    cleanup_on_start: bool = True
    enable_metrics: bool = False
    enable_encryption: bool = False
    encryption_key: Optional[str] = None  # base64-encoded 32-byte key
    max_key_length: int = 512
    circuit_breaker_threshold: int = 50
    circuit_breaker_window: int = 60


# === Storage Backend ===
class StorageBackend(ABC):
    @abstractmethod
    async def get(self, key: str) -> Optional[bytes]: ...
    @abstractmethod
    async def set(self, key: str, value: bytes, ttl: int) -> None: ...
    @abstractmethod
    async def cleanup_expired(self) -> int: ...
    @abstractmethod
    async def close(self) -> None: ...


class SQLiteStorage(StorageBackend):
    def __init__(self, db_path: str, config: CacheConfig):
        self.db_path = db_path
        self.config = config
        self._connection: Optional[aiosqlite.Connection] = None
        self._lock = asyncio.Lock()
        self._encryption_key = None

        if config.enable_encryption and config.encryption_key:
            if not HAS_CRYPTO:
                raise RuntimeError("Install 'cryptography' for encryption")
            import base64
            raw_key = base64.b64decode(config.encryption_key)
            if len(raw_key) != 32:
                raise ValueError("Encryption key must be 32 bytes (base64-encoded)")
            self._encryption_key = raw_key

    @staticmethod
    async def _init_db(conn: aiosqlite.Connection):
        await conn.execute("PRAGMA journal_mode=WAL")
        await conn.execute("PRAGMA synchronous=NORMAL")
        await conn.execute("PRAGMA cache_size=-10000")
        await conn.execute("PRAGMA temp_store=MEMORY")

        await conn.execute("""
            CREATE TABLE IF NOT EXISTS cache (
                key TEXT PRIMARY KEY,
                value BLOB NOT NULL,
                expires_at REAL NOT NULL
            )
        """)
        await conn.execute("CREATE INDEX IF NOT EXISTS idx_expires ON cache(expires_at)")

    @asynccontextmanager
    async def _get_connection(self):
        async with self._lock:
            if self._connection is None:
                self._connection = await aiosqlite.connect(
                    self.db_path,
                    detect_types=0,
                    isolation_level=None,
                    timeout=30.0,
                )
                await self._init_db(self._connection)
            yield self._connection

    def _encrypt(self, data: bytes) -> bytes:
        if not self._encryption_key:
            return data
        aesgcm = AESGCM(self._encryption_key)
        nonce = os.urandom(12)
        ct = aesgcm.encrypt(nonce, data, None)
        return nonce + ct

    def _decrypt(self, data: bytes) -> bytes:
        if not self._encryption_key:
            return data
        if len(data) < 12:
            raise ValueError("Invalid encrypted data")
        nonce, ct = data[:12], data[12:]
        aesgcm = AESGCM(self._encryption_key)
        return aesgcm.decrypt(nonce, ct, None)

    async def get(self, key: str) -> Optional[bytes]:
        now = time.time()
        async with self._get_connection() as conn:
            cursor = await conn.execute(
                "SELECT value FROM cache WHERE key = ? AND expires_at > ?", (key, now)
            )
            row = await cursor.fetchone()
            if row and row[0] is not None:
                return self._decrypt(row[0])
            return None

    async def set(self, key: str, value: bytes, ttl: int):
        expires_at = time.time() + ttl
        encrypted = self._encrypt(value)
        async with self._get_connection() as conn:
            await conn.execute(
                "INSERT OR REPLACE INTO cache (key, value, expires_at) VALUES (?, ?, ?)",
                (key, encrypted, expires_at)
            )
            await conn.commit()

    async def cleanup_expired(self) -> int:
        now = time.time()
        async with self._get_connection() as conn:
            cursor = await conn.execute("DELETE FROM cache WHERE expires_at <= ?", (now,))
            await conn.commit()
            return cursor.rowcount

    async def close(self):
        if self._connection:
            await self._connection.close()
            self._connection = None


# === Circuit Breaker ===
class CircuitBreaker:
    def __init__(self, failure_threshold: int, recovery_timeout: int):
        self.failure_threshold = failure_threshold
        self.recovery_timeout = recovery_timeout
        self.failure_count = 0
        self.last_failure_time = 0
        self.state = "CLOSED"
        self._lock = threading.Lock()

    def call_failed(self):
        with self._lock:
            self.failure_count += 1
            self.last_failure_time = time.time()
            if self.failure_count >= self.failure_threshold:
                self.state = "OPEN"

    def call_succeeded(self):
        with self._lock:
            self.failure_count = 0
            if self.state == "HALF_OPEN":
                self.state = "CLOSED"

    def can_execute(self) -> bool:
        with self._lock:
            if self.state == "CLOSED":
                return True
            if self.state == "OPEN":
                if time.time() - self.last_failure_time > self.recovery_timeout:
                    self.state = "HALF_OPEN"
                    return True
                return False
            return True


# === Main Cache ===
class AsyncCache:
    def __init__(self, config: CacheConfig):
        self.config = config
        self.storage = SQLiteStorage(config.db_path, config)
        self._l1_cache = TTLLRUCache(maxsize=config.l1_maxsize, ttl=config.l1_ttl)
        self._l1_lock = threading.RLock()
        self._circuit_breaker = CircuitBreaker(
            failure_threshold=config.circuit_breaker_threshold,
            recovery_timeout=config.circuit_breaker_window
        )
        self._tracer = trace.get_tracer(__name__)
        self._shutdown_event = asyncio.Event()
        self._gc_task: Optional[asyncio.Task] = None

        # Metrics
        self._metrics = None
        if config.enable_metrics:
            self._init_metrics(config.name)

        if config.vacuum_interval:
            self._gc_task = asyncio.create_task(self._maintenance_loop())

    @property
    def l1_cache(self) -> TTLLRUCache:
        return self._l1_cache

    @property
    def l1_lock(self) -> threading.RLock:
        return self._l1_lock

    def _init_metrics(self, cache_name: str):
        try:
            from prometheus_client import Counter, Histogram
            self._metrics = {
                "requests": Counter("cache_requests_total", "Cache requests", ["cache", "type"]),
                "duration": Histogram("cache_operation_duration_seconds", "Operation duration", ["cache", "operation"]),
                "errors": Counter("cache_errors_total", "Cache errors", ["cache", "error_type"]),
            }
        except ImportError:
            self._metrics = None

    async def _maintenance_loop(self):
        while not self._shutdown_event.is_set():
            # Спим по частям, чтобы быстро реагировать на shutdown
            slept = 0
            interval = self.config.vacuum_interval
            while slept < interval and not self._shutdown_event.is_set():
                sleep_step = min(1.0, interval - slept)  # проверяем каждую секунду
                await asyncio.sleep(sleep_step)
                slept += sleep_step

            if not self._shutdown_event.is_set():
                await self.storage.cleanup_expired()

    async def get(self, key: str) -> Optional[Any]:
        if not self._circuit_breaker.can_execute():
            logger.warning("circuit_breaker_open", key=key)
            if self._metrics:
                self._metrics["errors"].labels(error_type="circuit_breaker").inc()
            return None

        with self._tracer.start_as_current_span("cache.get", kind=SpanKind.CLIENT) as span:
            span.set_attribute("cache.key", key)
            start = time.perf_counter()

            try:
                with self._l1_lock:
                    if key in self._l1_cache:
                        span.set_attribute("cache.hit", "l1")
                        if self._metrics:
                            self._metrics["requests"].labels(type="l1_hit").inc()
                        return self._l1_cache[key]

                raw = await self.storage.get(key)
                if raw is None:
                    span.set_attribute("cache.hit", "miss")
                    if self._metrics:
                        self._metrics["requests"].labels(type="miss").inc()
                    return None

                value = pickle.loads(raw)
                with self._l1_lock:
                    self._l1_cache[key] = value
                span.set_attribute("cache.hit", "l2")
                if self._metrics:
                    self._metrics["requests"].labels(type="l2_hit").inc()
                self._circuit_breaker.call_succeeded()
                return value

            except Exception as e:
                self._circuit_breaker.call_failed()
                error_type = type(e).__name__
                logger.error("cache_get_error", key=key, error=error_type)
                if self._metrics:
                    self._metrics["errors"].labels(error_type=error_type).inc()
                span.record_exception(e)
                return None
            finally:
                if self._metrics:
                    self._metrics["duration"].labels(operation="get").observe(time.perf_counter() - start)

    async def set(self, key: str, value: Any, ttl: int):
        if not self._circuit_breaker.can_execute():
            return

        with self._tracer.start_as_current_span("cache.set", kind=SpanKind.CLIENT) as span:
            span.set_attribute("cache.key", key)
            start = time.perf_counter()

            try:
                pickled = pickle.dumps(value, protocol=pickle.HIGHEST_PROTOCOL)
                await self.storage.set(key, pickled, ttl)

                with self._l1_lock:
                    self._l1_cache[key] = value

                self._circuit_breaker.call_succeeded()
            except Exception as e:
                self._circuit_breaker.call_failed()
                error_type = type(e).__name__
                logger.error("cache_set_error", key=key, error=error_type)
                if self._metrics:
                    self._metrics["errors"].labels(error_type=error_type).inc()
                span.record_exception(e)
            finally:
                if self._metrics:
                    self._metrics["duration"].labels(operation="set").observe(time.perf_counter() - start)

    async def health_check(self) -> Dict[str, Any]:
        try:
            test_key = "health_" + secrets.token_hex(8)
            await self.set(test_key, "ok", 10)
            val = await self.get(test_key)
            healthy = (val == "ok")
        except Exception as e:
            healthy = False
            logger.error("health_check_failed", error=str(e))

        return {
            "status": "healthy" if healthy else "unhealthy",
            "l1_size": len(self._l1_cache),
            "circuit_breaker": self._circuit_breaker.state,
            "storage": "ok" if healthy else "failed"
        }

    async def graceful_shutdown(self):
        self._shutdown_event.set()
        if self._gc_task:
            try:
                await asyncio.wait_for(self._gc_task, timeout=5.0)
            except asyncio.TimeoutError:
                logger.warning("gc_task_timeout")
        await self.storage.close()

    def get_metrics_text(self) -> str:
        return generate_latest().decode('utf-8')



import pytest
import asyncio
import inspect
from cachka import cached, cache_registry, CacheConfig


# Глобальный класс для тестов (чтобы pickle мог его сериализовать)
class TestService:
    """Глобальный класс для тестов ignore_self"""
    def __init__(self, name):
        self.name = name
    
    def get_data(self, key: str):
        return f"data_{key}_{self.name}"


class TestDecoratorAsync:
    """Тесты декоратора для async функций"""

    @pytest.fixture(autouse=True)
    async def setup_cache(self):
        # Сбрасываем перед инициализацией
        if cache_registry.is_initialized():
            try:
                await cache_registry.shutdown()
            except:
                pass
            cache_registry.reset()
        
        config = CacheConfig(
            db_path=":memory:",
            vacuum_interval=None,
            cleanup_on_start=False
        )
        cache_registry.initialize(config)
        yield
        if cache_registry.is_initialized():
            try:
                await cache_registry.shutdown()
            except:
                pass
            cache_registry.reset()

    @pytest.mark.asyncio
    async def test_async_function_caching(self):
        """Кэширование async функций"""
        call_count = [0]
        
        @cached(ttl=60)
        async def fetch_data(key: str):
            call_count[0] += 1
            await asyncio.sleep(0.01)
            return f"data_{key}"
        
        result1 = await fetch_data("test")
        assert result1 == "data_test"
        assert call_count[0] == 1
        
        result2 = await fetch_data("test")
        assert result2 == "data_test"
        assert call_count[0] == 1  # Не вызвалась снова

    @pytest.mark.asyncio
    async def test_async_function_cache_hit(self):
        """Попадание в кэш"""
        call_count = [0]
        
        @cached(ttl=60)
        async def compute(x: int):
            call_count[0] += 1
            return x * 2
        
        await compute(5)
        await compute(5)  # Из кэша
        assert call_count[0] == 1

    @pytest.mark.asyncio
    async def test_async_function_different_args(self):
        """Разные аргументы = разные ключи"""
        call_count = [0]
        
        @cached(ttl=60)
        async def compute(x: int):
            call_count[0] += 1
            return x * 2
        
        await compute(5)
        await compute(10)  # Другой аргумент
        assert call_count[0] == 2

    @pytest.mark.asyncio
    async def test_async_function_kwargs(self):
        """Работа с kwargs"""
        call_count = [0]
        
        @cached(ttl=60)
        async def compute(x: int, multiplier: int = 2):
            call_count[0] += 1
            return x * multiplier
        
        await compute(5, multiplier=3)
        await compute(5, multiplier=3)  # Из кэша
        assert call_count[0] == 1

    @pytest.mark.asyncio
    async def test_async_function_ttl(self):
        """Соблюдение TTL"""
        call_count = [0]
        
        @cached(ttl=1)
        async def compute(x: int):
            call_count[0] += 1
            return x * 2
        
        await compute(5)
        await compute(5)  # Из кэша
        assert call_count[0] == 1
        
        # Ждем истечения TTL
        await asyncio.sleep(1.2)
        # Очищаем кэш вручную для теста
        cache = cache_registry.get()
        await cache.storage.cleanup_expired()
        with cache.l1_lock:
            cache.l1_cache.cleanup()
            # Удаляем ключ из L1, если он там есть
            from cachka.utils import make_cache_key
            cache_key = make_cache_key("compute", (5,), {})
            if cache_key in cache.l1_cache:
                cache.l1_cache.delete(cache_key)
        
        await compute(5)  # TTL истек, должна вызваться снова
        assert call_count[0] == 2


class TestDecoratorSync:
    """Тесты декоратора для sync функций"""

    @pytest.fixture(autouse=True)
    async def setup_cache(self):
        # Сбрасываем перед инициализацией
        if cache_registry.is_initialized():
            try:
                await cache_registry.shutdown()
            except:
                pass
            cache_registry.reset()
        
        config = CacheConfig(
            db_path=":memory:",
            vacuum_interval=None,
            cleanup_on_start=False
        )
        cache_registry.initialize(config)
        yield
        if cache_registry.is_initialized():
            try:
                await cache_registry.shutdown()
            except:
                pass
            cache_registry.reset()

    def test_sync_function_caching(self):
        """Кэширование sync функций"""
        call_count = [0]
        
        @cached(ttl=60)
        def compute(x: int):
            call_count[0] += 1
            return x * 2
        
        result1 = compute(5)
        assert result1 == 10
        assert call_count[0] == 1
        
        result2 = compute(5)
        assert result2 == 10
        assert call_count[0] == 1  # Не вызвалась снова

    def test_sync_function_cache_hit(self):
        """Попадание в кэш"""
        call_count = [0]
        
        @cached(ttl=60)
        def compute(x: int):
            call_count[0] += 1
            return x * 2
        
        compute(5)
        compute(5)  # Из кэша
        assert call_count[0] == 1

    def test_sync_function_l1_cache(self):
        """Использование L1 кэша"""
        @cached(ttl=60)
        def compute(x: int):
            return x * 2
        
        compute(5)
        # Второй вызов должен использовать L1
        result = compute(5)
        assert result == 10


class TestDecoratorIgnoreSelf:
    """Тесты ignore_self"""

    @pytest.fixture(autouse=True)
    async def setup_cache(self):
        # Сбрасываем перед инициализацией
        if cache_registry.is_initialized():
            try:
                await cache_registry.shutdown()
            except:
                pass
            cache_registry.reset()
        
        config = CacheConfig(
            db_path=":memory:",
            vacuum_interval=None,
            cleanup_on_start=False
        )
        cache_registry.initialize(config)
        yield
        if cache_registry.is_initialized():
            try:
                await cache_registry.shutdown()
            except:
                pass
            cache_registry.reset()

    @pytest.mark.asyncio
    async def test_ignore_self_true(self):
        """Игнорирование self в ключе"""
        call_count = [0]
        
        # Используем глобальный класс для pickle
        class ServiceIgnoreSelf(TestService):
            @cached(ttl=60, ignore_self=True)
            async def get_data(self, key: str):
                call_count[0] += 1
                return f"data_{key}"
        
        service1 = ServiceIgnoreSelf("service1")
        service2 = ServiceIgnoreSelf("service2")
        
        result1 = await service1.get_data("test")
        result2 = await service2.get_data("test")  # Должно быть из кэша
        
        assert result1 == result2
        assert call_count[0] == 1  # Вызвалась только один раз

    def test_ignore_self_false(self):
        """Включение self в ключ"""
        call_count = [0]
        
        # Используем глобальный класс для pickle
        class ServiceWithCount(TestService):
            @cached(ttl=60, ignore_self=False)
            def get_data(self, key: str):
                call_count[0] += 1
                return super().get_data(key)
        
        service1 = ServiceWithCount("service1")
        service2 = ServiceWithCount("service2")
        
        service1.get_data("test")
        service2.get_data("test")  # Разные экземпляры = разные ключи
        
        assert call_count[0] == 2  # Вызвалась дважды


class TestDecoratorMetadata:
    """Тесты сохранения метаданных функции"""

    @pytest.fixture(autouse=True)
    async def setup_cache(self):
        # Сбрасываем перед инициализацией
        if cache_registry.is_initialized():
            try:
                await cache_registry.shutdown()
            except:
                pass
            cache_registry.reset()
        
        config = CacheConfig(
            db_path=":memory:",
            vacuum_interval=None,
            cleanup_on_start=False
        )
        cache_registry.initialize(config)
        yield
        if cache_registry.is_initialized():
            try:
                await cache_registry.shutdown()
            except:
                pass
            cache_registry.reset()

    @pytest.mark.asyncio
    async def test_preserves_function_name(self):
        """Сохранение __name__"""
        @cached(ttl=60)
        async def my_function(x: int):
            """Test function"""
            return x * 2
        
        assert my_function.__name__ == "my_function"

    @pytest.mark.asyncio
    async def test_preserves_function_doc(self):
        """Сохранение __doc__"""
        @cached(ttl=60)
        async def my_function(x: int):
            """Test function docstring"""
            return x * 2
        
        assert my_function.__doc__ == "Test function docstring"

    @pytest.mark.asyncio
    async def test_preserves_function_annotations(self):
        """Сохранение __annotations__ (для FastAPI)"""
        @cached(ttl=60)
        async def my_function(x: int, y: str = "default") -> dict:
            return {"x": x, "y": y}
        
        assert "x" in my_function.__annotations__
        assert "y" in my_function.__annotations__
        assert "return" in my_function.__annotations__
        assert my_function.__annotations__["x"] == int
        assert my_function.__annotations__["return"] == dict

    def test_preserves_function_signature(self):
        """Сохранение сигнатуры"""
        @cached(ttl=60)
        def my_function(x: int, y: str = "default") -> dict:
            return {"x": x, "y": y}
        
        sig = inspect.signature(my_function)
        assert "x" in sig.parameters
        assert "y" in sig.parameters
        assert sig.return_annotation == dict


class TestDecoratorEdgeCases:
    """Edge cases декоратора"""

    @pytest.fixture(autouse=True)
    async def setup_cache(self):
        # Сбрасываем перед инициализацией
        if cache_registry.is_initialized():
            try:
                await cache_registry.shutdown()
            except:
                pass
            cache_registry.reset()
        
        config = CacheConfig(
            db_path=":memory:",
            vacuum_interval=None,
            cleanup_on_start=False
        )
        cache_registry.initialize(config)
        yield
        if cache_registry.is_initialized():
            try:
                await cache_registry.shutdown()
            except:
                pass
            cache_registry.reset()

    @pytest.mark.asyncio
    async def test_decorator_with_no_args(self):
        """Декоратор без аргументов"""
        @cached()
        async def compute(x: int):
            return x * 2
        
        result = await compute(5)
        assert result == 10

    @pytest.mark.asyncio
    async def test_decorator_with_kwargs_only(self):
        """Только kwargs"""
        @cached(ttl=60)
        async def compute(x: int, multiplier: int = 2):
            return x * multiplier
        
        result = await compute(5, multiplier=3)
        assert result == 15

    @pytest.mark.asyncio
    async def test_decorator_with_args_and_kwargs(self):
        """Args и kwargs"""
        @cached(ttl=60)
        async def compute(x: int, y: int, multiplier: int = 2):
            return (x + y) * multiplier
        
        result = await compute(5, 10, multiplier=3)
        assert result == 45

    def test_decorator_recursive_function(self):
        """Рекурсивные функции"""
        call_count = [0]
        
        @cached(ttl=60)
        def fib(n: int):
            call_count[0] += 1
            if n < 2:
                return n
            return fib(n - 1) + fib(n - 2)
        
        result = fib(10)
        assert result == 55
        # Проверяем, что функция вызывалась (не зациклилась)
        assert call_count[0] > 0
        # Проверяем, что кэширование работает - второй вызов должен использовать кэш
        call_count[0] = 0
        result2 = fib(10)
        assert result2 == 55
        assert call_count[0] == 0  # Не должна вызываться снова




import hashlib
import pickle
from typing import Tuple

def make_cache_key(func_name: str, args: Tuple, kwargs: dict) -> str:
    key_data = (func_name, args, tuple(sorted(kwargs.items())))
    return hashlib.sha256(
        pickle.dumps(key_data, protocol=pickle.HIGHEST_PROTOCOL)
    ).hexdigest()